ZipperEx.Zipable protocol (ZipperEx v0.2.0) View Source
Zipable protocol used by ZipperEx.
See ZipperEx for an example.
Link to this section Summary
Functions
Returns true if the given node is a branch.
Returns the children of the given node.
Creates a node from the given node and children.
Link to this section Types
Specs
t() :: term()
Link to this section Functions
Specs
branch?(ZipperEx.t()) :: boolean()
Returns true if the given node is a branch.
Specs
children(ZipperEx.t()) :: [ZipperEx.tree()]
Returns the children of the given node.
Specs
make_node(ZipperEx.t(), [ZipperEx.tree()]) :: ZipperEx.t()
Creates a node from the given node and children.